/***********************************************************************
 3NEWDE - NewDE cipher in simple variant EDE mode

 Key lengh: 192 bit, block size: 64 bit

 Implemented by Alexander Myasnikow

Web: www.darksoftware.narod.ru

************************************************************************/

#include <string.h>

typedef unsigned long int u32;
typedef signed long int s32;
typedef unsigned char u8;


/* internal functions that no one else needs to know about */

static void ffunc (u32 *, u32 *);
static void scrunch (u8 *, u32 *);
static void unscrun (u32 *, u8 *);

/* internal (active) key registers */

static u32 KnLE[2][32];
static u32 KnLD[2][32];


void __stdcall __export
setup (u8 * key)
{
  register s32 i, n;
  register u32 *tofill;
  u32 mkr[16];
  u8 hexdes[2][16];


  for (n = 0; n != 2; n++)
    {


      for (i = 0; i != 16; i++)
	{

	  hexdes[n][i] = (key[(n << 4) + i] >> 4) & 0x0f;
	  hexdes[n][i + 1] = (key[(n << 4) + i]) & 0x0f;
	}

      for (i = 0; i != 32; i++)
	{
	  KnLE[n][i] = KnLD[n][i] = 0;
	}

      for (i = 0; i != 16; i++)
	{
	  mkr[i] = hexdes[n][i];
	}

      for (i = 0; i != 16; i++)
	{
	  tofill = &KnLE[n][i << 1];

	  *tofill = ((mkr[(i + 13) & 0x0f] << 2) | (mkr[i] >> 2));
	  *tofill |=
	    ((mkr[(i + 6) & 0x0f] << 2) | (mkr[(i + 11) & 0x0f] >> 2)) << 8;
	  *tofill |=
	    ((mkr[(i + 10) & 0x0f] << 2) | (mkr[(i + 3) & 0x0f] >> 2)) << 16;
	  *tofill++ |=
	    ((mkr[(i + 1) & 0x0f] << 2) | (mkr[(i + 8) & 0x0f] >> 2)) << 24;
	  *tofill = (((mkr[i] << 4) | mkr[(i + 9) & 0x0f]) & 0x3fL);
	  *tofill |=
	    (((mkr[(i + 11) & 0x0f] << 4) | mkr[(i + 2) & 0x0f]) & 0x3fL) <<
	    8;
	  *tofill |=
	    (((mkr[(i + 3) & 0x0f] << 4) | mkr[(i + 14) & 0x0f]) & 0x3fL) <<
	    16;
	  *tofill |=
	    (((mkr[(i + 8) & 0x0f] << 4) | mkr[(i + 5) & 0x0f]) & 0x3fL) <<
	    24;
	}



      for (i = 0; i < 16; i++)
	{
	  tofill = &KnLD[n][(15 - i) << 1];

	  *tofill = ((mkr[(i + 13) & 0x0f] << 2) | (mkr[i] >> 2));
	  *tofill |=
	    ((mkr[(i + 6) & 0x0f] << 2) | (mkr[(i + 11) & 0x0f] >> 2)) << 8;
	  *tofill |=
	    ((mkr[(i + 10) & 0x0f] << 2) | (mkr[(i + 3) & 0x0f] >> 2)) << 16;
	  *tofill++ |=
	    ((mkr[(i + 1) & 0x0f] << 2) | (mkr[(i + 8) & 0x0f] >> 2)) << 24;
	  *tofill = (((mkr[i] << 4) | mkr[(i + 9) & 0x0f]) & 0x3fL);
	  *tofill |=
	    (((mkr[(i + 11) & 0x0f] << 4) | mkr[(i + 2) & 0x0f]) & 0x3fL) <<
	    8;
	  *tofill |=
	    (((mkr[(i + 3) & 0x0f] << 4) | mkr[(i + 14) & 0x0f]) & 0x3fL) <<
	    16;
	  *tofill |=
	    (((mkr[(i + 8) & 0x0f] << 4) | mkr[(i + 5) & 0x0f]) & 0x3fL) <<
	    24;

	}



    }

  n = 4;

}


void __stdcall __export
crypt (u8 * inblock)
{
  register u32 *work;
  u32 buff[2], i;
  u8 outblock[8];

  work = buff;
  scrunch (inblock, work);
  ffunc (work, KnLE[0]);
  ffunc (work, KnLD[1]);
  ffunc (work, KnLE[1]);
  unscrun (work, &outblock[0]);

  for (i = 0; i != 8; i++)
    {
      inblock[i] = outblock[i];
    }

}


void __stdcall __export
decrypt (u8 * inblock)
{
  register u32 *work;
  u32 buff[2], i;
  u8 outblock[8];

  work = buff;
  scrunch (inblock, work);
  ffunc (work, KnLD[1]);
  ffunc (work, KnLE[1]);
  ffunc (work, KnLD[0]);
  unscrun (work, &outblock[0]);

  for (i = 0; i != 8; i++)
    {
      inblock[i] = outblock[i];
    }

}


/* Portable data packing and unpacking */

static void
scrunch (outof, into)
     register u8 *outof;
     register u32 *into;
{
  *into = (*outof++ & 0xffL) << 24;
  *into |= (*outof++ & 0xffL) << 16;
  *into |= (*outof++ & 0xffL) << 8;
  *into++ |= (*outof++ & 0xffL);
  *into = (*outof++ & 0xffL) << 24;
  *into |= (*outof++ & 0xffL) << 16;
  *into |= (*outof++ & 0xffL) << 8;
  *into |= (*outof & 0xffL);
  return;
}

static void
unscrun (outof, into)
     register u32 *outof;
     register u8 *into;
{
  *into++ = (u8) ((*outof >> 24) & 0xffL);
  *into++ = (u8) ((*outof >> 16) & 0xffL);
  *into++ = (u8) ((*outof >> 8) & 0xffL);
  *into++ = (u8) (*outof++ & 0xffL);
  *into++ = (u8) ((*outof >> 24) & 0xffL);
  *into++ = (u8) ((*outof >> 16) & 0xffL);
  *into++ = (u8) ((*outof >> 8) & 0xffL);
  *into = (u8) (*outof & 0xffL);
  return;
}

/* The combined pre-computed Substitution and Permutation boxes */

static u32 SP1[64] = {
  0x01010400L, 0x00000000L, 0x00010000L, 0x01010404L,
  0x01010004L, 0x00010404L, 0x00000004L, 0x00010000L,
  0x00000400L, 0x01010400L, 0x01010404L, 0x00000400L,
  0x01000404L, 0x01010004L, 0x01000000L, 0x00000004L,
  0x00000404L, 0x01000400L, 0x01000400L, 0x00010400L,
  0x00010400L, 0x01010000L, 0x01010000L, 0x01000404L,
  0x00010004L, 0x01000004L, 0x01000004L, 0x00010004L,
  0x00000000L, 0x00000404L, 0x00010404L, 0x01000000L,
  0x00010000L, 0x01010404L, 0x00000004L, 0x01010000L,
  0x01010400L, 0x01000000L, 0x01000000L, 0x00000400L,
  0x01010004L, 0x00010000L, 0x00010400L, 0x01000004L,
  0x00000400L, 0x00000004L, 0x01000404L, 0x00010404L,
  0x01010404L, 0x00010004L, 0x01010000L, 0x01000404L,
  0x01000004L, 0x00000404L, 0x00010404L, 0x01010400L,
  0x00000404L, 0x01000400L, 0x01000400L, 0x00000000L,
  0x00010004L, 0x00010400L, 0x00000000L, 0x01010004L
};

static u32 SP2[64] = {
  0x80108020L, 0x80008000L, 0x00008000L, 0x00108020L,
  0x00100000L, 0x00000020L, 0x80100020L, 0x80008020L,
  0x80000020L, 0x80108020L, 0x80108000L, 0x80000000L,
  0x80008000L, 0x00100000L, 0x00000020L, 0x80100020L,
  0x00108000L, 0x00100020L, 0x80008020L, 0x00000000L,
  0x80000000L, 0x00008000L, 0x00108020L, 0x80100000L,
  0x00100020L, 0x80000020L, 0x00000000L, 0x00108000L,
  0x00008020L, 0x80108000L, 0x80100000L, 0x00008020L,
  0x00000000L, 0x00108020L, 0x80100020L, 0x00100000L,
  0x80008020L, 0x80100000L, 0x80108000L, 0x00008000L,
  0x80100000L, 0x80008000L, 0x00000020L, 0x80108020L,
  0x00108020L, 0x00000020L, 0x00008000L, 0x80000000L,
  0x00008020L, 0x80108000L, 0x00100000L, 0x80000020L,
  0x00100020L, 0x80008020L, 0x80000020L, 0x00100020L,
  0x00108000L, 0x00000000L, 0x80008000L, 0x00008020L,
  0x80000000L, 0x80100020L, 0x80108020L, 0x00108000L
};

static u32 SP3[64] = {
  0x00000208L, 0x08020200L, 0x00000000L, 0x08020008L,
  0x08000200L, 0x00000000L, 0x00020208L, 0x08000200L,
  0x00020008L, 0x08000008L, 0x08000008L, 0x00020000L,
  0x08020208L, 0x00020008L, 0x08020000L, 0x00000208L,
  0x08000000L, 0x00000008L, 0x08020200L, 0x00000200L,
  0x00020200L, 0x08020000L, 0x08020008L, 0x00020208L,
  0x08000208L, 0x00020200L, 0x00020000L, 0x08000208L,
  0x00000008L, 0x08020208L, 0x00000200L, 0x08000000L,
  0x08020200L, 0x08000000L, 0x00020008L, 0x00000208L,
  0x00020000L, 0x08020200L, 0x08000200L, 0x00000000L,
  0x00000200L, 0x00020008L, 0x08020208L, 0x08000200L,
  0x08000008L, 0x00000200L, 0x00000000L, 0x08020008L,
  0x08000208L, 0x00020000L, 0x08000000L, 0x08020208L,
  0x00000008L, 0x00020208L, 0x00020200L, 0x08000008L,
  0x08020000L, 0x08000208L, 0x00000208L, 0x08020000L,
  0x00020208L, 0x00000008L, 0x08020008L, 0x00020200L
};

static u32 SP4[64] = {
  0x00802001L, 0x00002081L, 0x00002081L, 0x00000080L,
  0x00802080L, 0x00800081L, 0x00800001L, 0x00002001L,
  0x00000000L, 0x00802000L, 0x00802000L, 0x00802081L,
  0x00000081L, 0x00000000L, 0x00800080L, 0x00800001L,
  0x00000001L, 0x00002000L, 0x00800000L, 0x00802001L,
  0x00000080L, 0x00800000L, 0x00002001L, 0x00002080L,
  0x00800081L, 0x00000001L, 0x00002080L, 0x00800080L,
  0x00002000L, 0x00802080L, 0x00802081L, 0x00000081L,
  0x00800080L, 0x00800001L, 0x00802000L, 0x00802081L,
  0x00000081L, 0x00000000L, 0x00000000L, 0x00802000L,
  0x00002080L, 0x00800080L, 0x00800081L, 0x00000001L,
  0x00802001L, 0x00002081L, 0x00002081L, 0x00000080L,
  0x00802081L, 0x00000081L, 0x00000001L, 0x00002000L,
  0x00800001L, 0x00002001L, 0x00802080L, 0x00800081L,
  0x00002001L, 0x00002080L, 0x00800000L, 0x00802001L,
  0x00000080L, 0x00800000L, 0x00002000L, 0x00802080L
};

static u32 SP5[64] = {
  0x00000100L, 0x02080100L, 0x02080000L, 0x42000100L,
  0x00080000L, 0x00000100L, 0x40000000L, 0x02080000L,
  0x40080100L, 0x00080000L, 0x02000100L, 0x40080100L,
  0x42000100L, 0x42080000L, 0x00080100L, 0x40000000L,
  0x02000000L, 0x40080000L, 0x40080000L, 0x00000000L,
  0x40000100L, 0x42080100L, 0x42080100L, 0x02000100L,
  0x42080000L, 0x40000100L, 0x00000000L, 0x42000000L,
  0x02080100L, 0x02000000L, 0x42000000L, 0x00080100L,
  0x00080000L, 0x42000100L, 0x00000100L, 0x02000000L,
  0x40000000L, 0x02080000L, 0x42000100L, 0x40080100L,
  0x02000100L, 0x40000000L, 0x42080000L, 0x02080100L,
  0x40080100L, 0x00000100L, 0x02000000L, 0x42080000L,
  0x42080100L, 0x00080100L, 0x42000000L, 0x42080100L,
  0x02080000L, 0x00000000L, 0x40080000L, 0x42000000L,
  0x00080100L, 0x02000100L, 0x40000100L, 0x00080000L,
  0x00000000L, 0x40080000L, 0x02080100L, 0x40000100L
};

static u32 SP6[64] = {
  0x20000010L, 0x20400000L, 0x00004000L, 0x20404010L,
  0x20400000L, 0x00000010L, 0x20404010L, 0x00400000L,
  0x20004000L, 0x00404010L, 0x00400000L, 0x20000010L,
  0x00400010L, 0x20004000L, 0x20000000L, 0x00004010L,
  0x00000000L, 0x00400010L, 0x20004010L, 0x00004000L,
  0x00404000L, 0x20004010L, 0x00000010L, 0x20400010L,
  0x20400010L, 0x00000000L, 0x00404010L, 0x20404000L,
  0x00004010L, 0x00404000L, 0x20404000L, 0x20000000L,
  0x20004000L, 0x00000010L, 0x20400010L, 0x00404000L,
  0x20404010L, 0x00400000L, 0x00004010L, 0x20000010L,
  0x00400000L, 0x20004000L, 0x20000000L, 0x00004010L,
  0x20000010L, 0x20404010L, 0x00404000L, 0x20400000L,
  0x00404010L, 0x20404000L, 0x00000000L, 0x20400010L,
  0x00000010L, 0x00004000L, 0x20400000L, 0x00404010L,
  0x00004000L, 0x00400010L, 0x20004010L, 0x00000000L,
  0x20404000L, 0x20000000L, 0x00400010L, 0x20004010L
};

static u32 SP7[64] = {
  0x00200000L, 0x04200002L, 0x04000802L, 0x00000000L,
  0x00000800L, 0x04000802L, 0x00200802L, 0x04200800L,
  0x04200802L, 0x00200000L, 0x00000000L, 0x04000002L,
  0x00000002L, 0x04000000L, 0x04200002L, 0x00000802L,
  0x04000800L, 0x00200802L, 0x00200002L, 0x04000800L,
  0x04000002L, 0x04200000L, 0x04200800L, 0x00200002L,
  0x04200000L, 0x00000800L, 0x00000802L, 0x04200802L,
  0x00200800L, 0x00000002L, 0x04000000L, 0x00200800L,
  0x04000000L, 0x00200800L, 0x00200000L, 0x04000802L,
  0x04000802L, 0x04200002L, 0x04200002L, 0x00000002L,
  0x00200002L, 0x04000000L, 0x04000800L, 0x00200000L,
  0x04200800L, 0x00000802L, 0x00200802L, 0x04200800L,
  0x00000802L, 0x04000002L, 0x04200802L, 0x04200000L,
  0x00200800L, 0x00000000L, 0x00000002L, 0x04200802L,
  0x00000000L, 0x00200802L, 0x04200000L, 0x00000800L,
  0x04000002L, 0x04000800L, 0x00000800L, 0x00200002L
};

static u32 SP8[64] = {
  0x10001040L, 0x00001000L, 0x00040000L, 0x10041040L,
  0x10000000L, 0x10001040L, 0x00000040L, 0x10000000L,
  0x00040040L, 0x10040000L, 0x10041040L, 0x00041000L,
  0x10041000L, 0x00041040L, 0x00001000L, 0x00000040L,
  0x10040000L, 0x10000040L, 0x10001000L, 0x00001040L,
  0x00041000L, 0x00040040L, 0x10040040L, 0x10041000L,
  0x00001040L, 0x00000000L, 0x00000000L, 0x10040040L,
  0x10000040L, 0x10001000L, 0x00041040L, 0x00040000L,
  0x00041040L, 0x00040000L, 0x10041000L, 0x00001000L,
  0x00000040L, 0x10040040L, 0x00001000L, 0x00041040L,
  0x10001000L, 0x00000040L, 0x10000040L, 0x10040000L,
  0x10040040L, 0x10000000L, 0x00040000L, 0x10001040L,
  0x00000000L, 0x10041040L, 0x00040040L, 0x10000040L,
  0x10040000L, 0x10001000L, 0x10001040L, 0x00000000L,
  0x10041040L, 0x00041000L, 0x00041000L, 0x00001040L,
  0x00001040L, 0x00040040L, 0x10000000L, 0x10041000L
};

/* The central bit-crunching f-function */

static void
ffunc (block, keys)
     register u32 *block, *keys;
{
  register u32 fval, work, right, leftt;
  register s32 round;

  leftt = ((block[0] << 1) | ((block[0] >> 31) & 1L)) & 0xffffffffL;
  right = ((block[1] << 1) | ((block[1] >> 31) & 1L)) & 0xffffffffL;
  for (round = 0; round < 8; round++)
    {
      work = ((right << 28) | (right >> 4)) ^ *keys++;
      fval = SP7[work & 0x3fL];
      fval |= SP5[(work >> 8) & 0x3fL];
      fval |= SP3[(work >> 16) & 0x3fL];
      fval |= SP1[(work >> 24) & 0x3fL];
      work = right ^ *keys++;
      fval |= SP8[work & 0x3fL];
      fval |= SP6[(work >> 8) & 0x3fL];
      fval |= SP4[(work >> 16) & 0x3fL];
      fval |= SP2[(work >> 24) & 0x3fL];
      leftt ^= fval;
      work = ((leftt << 28) | (leftt >> 4)) ^ *keys++;
      fval = SP7[work & 0x3fL];
      fval |= SP5[(work >> 8) & 0x3fL];
      fval |= SP3[(work >> 16) & 0x3fL];
      fval |= SP1[(work >> 24) & 0x3fL];
      work = leftt ^ *keys++;
      fval |= SP8[work & 0x3fL];
      fval |= SP6[(work >> 8) & 0x3fL];
      fval |= SP4[(work >> 16) & 0x3fL];
      fval |= SP2[(work >> 24) & 0x3fL];
      right ^= fval;
    }
  *block++ = (right << 31) | (right >> 1);
  *block = (leftt << 31) | (leftt >> 1);
  return;
}







u32 __stdcall __export
getblocksize ()
{
  return 64;
}

u32 __stdcall __export
getkeysize ()
{
  return 192;
}

void __stdcall __export
getciphername (u8 * p)
{
  strcpy (p, "3NewDE");
}
